//! Directory synchronized handshake
//! 1-RTT connection
//!
//! The simplest, fastest handshake supported by Fenrir
//! Downside: It does not offer protection from DDos,
//! no perfect forward secrecy
//!
//! To grant a form of perfect forward secrecy, the server should periodically
//! change the DNSSEC public/private keys

use crate::enc::{
    sym::{NonceLen, TagLen},
    Random,
};

pub mod req;
pub mod resp;

// TODO: merge with crate::enc::sym::Nonce
/// random nonce
#[derive(Debug, Clone, Copy, PartialEq)]
pub struct Nonce(pub(crate) [u8; 16]);

impl Nonce {
    /// Create a new random Nonce
    pub fn new(rnd: &Random) -> Self {
        use ::core::mem::MaybeUninit;
        let mut out: MaybeUninit<[u8; 16]>;
        #[allow(unsafe_code)]
        unsafe {
            out = MaybeUninit::uninit();
            let _ = rnd.fill(out.assume_init_mut());
            Self(out.assume_init())
        }
    }
    /// Length of the serialized Nonce
    pub const fn len() -> usize {
        16
    }
}
impl From<&[u8; 16]> for Nonce {
    fn from(raw: &[u8; 16]) -> Self {
        Self(raw.clone())
    }
}

/// Parsed handshake
#[derive(Debug, Clone, PartialEq)]
pub enum DirSync {
    /// Directory synchronized handshake: client request
    Req(req::Req),
    /// Directory synchronized handshake: server response
    Resp(resp::Resp),
}

impl DirSync {
    /// actual length of the dirsync handshake data
    pub fn len(&self, head_len: NonceLen, tag_len: TagLen) -> usize {
        match self {
            DirSync::Req(req) => req.len(),
            DirSync::Resp(resp) => resp.len(head_len, tag_len),
        }
    }
    /// Serialize into raw bytes
    /// NOTE: assumes that there is exactly asa much buffer as needed
    pub fn serialize(
        &self,
        head_len: NonceLen,
        tag_len: TagLen,
        out: &mut [u8],
    ) {
        match self {
            DirSync::Req(req) => req.serialize(head_len, tag_len, out),
            DirSync::Resp(resp) => resp.serialize(head_len, tag_len, out),
        }
    }
}
