//! Connection handling and send/receive queues

pub mod handshake;
pub mod packet;
pub mod socket;
pub mod stream;

use ::core::num::Wrapping;
use ::std::{
    collections::{BTreeMap, HashMap, VecDeque},
    vec::Vec,
};

pub use crate::connection::{handshake::Handshake, packet::Packet};

use crate::{
    connection::{socket::UdpClient, stream::StreamData},
    dnssec,
    enc::{
        self,
        asym::PubKey,
        hkdf::Hkdf,
        sym::{self, CipherRecv, CipherSend},
        Random,
    },
    inner::{worker, ThreadTracker},
};

/// Connection errors
#[derive(::thiserror::Error, Debug, Copy, Clone)]
pub enum Error {
    /// Can't decrypt packet
    #[error("Decrypt error: {0}")]
    Decrypt(#[from] enc::Error),
    /// Error in parsing a packet realated to the connection
    #[error("Chunk parsing: {0}")]
    Parse(#[from] stream::Error),
    /// No such Connection
    #[error("No suck connection")]
    NoSuchConnection,
    /// No such Stream
    #[error("No suck Stream")]
    NoSuchStream,
}

/// Fenrir Connection ID
///
/// 0 is special as it represents the handshake  
/// Connection IDs are to be considered u64 little endian
#[derive(Debug, Copy, Clone, PartialEq, Eq, PartialOrd, Ord)]
pub enum ID {
    /// Connection id 0 represent the handshake
    Handshake,
    /// Non-zero id can represent any connection
    ID(::core::num::NonZeroU64),
}

impl ID {
    /// Set the conenction id to handshake
    pub fn new_handshake() -> Self {
        Self::Handshake
    }
    /// New id from u64. PLZ NON ZERO
    pub(crate) fn new_u64(raw: u64) -> Self {
        #[allow(unsafe_code)]
        unsafe {
            ID::ID(::core::num::NonZeroU64::new_unchecked(raw))
        }
    }
    pub(crate) fn as_u64(&self) -> u64 {
        match self {
            ID::Handshake => 0,
            ID::ID(id) => id.get(),
        }
    }
    /// New random service ID
    pub fn new_rand(rand: &Random) -> Self {
        let mut raw = [0; 8];
        let mut num = 0;
        while num == 0 {
            rand.fill(&mut raw);
            num = u64::from_le_bytes(raw);
        }
        #[allow(unsafe_code)]
        unsafe {
            ID::ID(::core::num::NonZeroU64::new_unchecked(num))
        }
    }
    /// Quick check to know if this is an handshake
    pub fn is_handshake(&self) -> bool {
        *self == ID::Handshake
    }
    /// length if the connection ID in bytes
    pub const fn len() -> usize {
        8
    }
    /// write the ID to a buffer
    pub fn serialize(&self, out: &mut [u8]) {
        match self {
            ID::Handshake => out[..8].copy_from_slice(&[0; 8]),
            ID::ID(id) => out[..8].copy_from_slice(&id.get().to_le_bytes()),
        }
    }
}

impl From<u64> for ID {
    fn from(raw: u64) -> Self {
        if raw == 0 {
            ID::Handshake
        } else {
            #[allow(unsafe_code)]
            unsafe {
                ID::ID(::core::num::NonZeroU64::new_unchecked(raw))
            }
        }
    }
}

impl From<[u8; 8]> for ID {
    fn from(raw: [u8; 8]) -> Self {
        let raw_u64 = u64::from_le_bytes(raw);
        raw_u64.into()
    }
}

/// strong typedef for receiving connection id
#[derive(Debug, Copy, Clone, PartialEq, Eq, PartialOrd, Ord)]
pub struct IDRecv(pub ID);
/// strong typedef for sending connection id
#[derive(Debug, Copy, Clone, PartialEq, Eq, PartialOrd, Ord)]
pub struct IDSend(pub ID);

/// Version of the fenrir protocol in use
#[derive(::num_derive::FromPrimitive, Debug, Copy, Clone, PartialEq)]
#[repr(u8)]
pub enum ProtocolVersion {
    /// First Fenrir Protocol Version
    V0 = 0,
}
impl ProtocolVersion {
    /// actual length of the protocol version field
    pub const fn len() -> usize {
        1
    }
    /// Serialize into raw bytes
    pub fn serialize(&self, out: &mut u8) {
        *out = *self as u8;
    }
}

/// Connection tracking id. Set by the user
#[derive(Debug, Clone, Copy, Eq, PartialEq, Ord, PartialOrd)]
pub struct UserTracker(pub ::core::num::NonZeroU64);

/// Unique tracker of connections
#[derive(Debug, Clone, Copy, Eq, PartialEq, Ord, PartialOrd)]
pub struct LibTracker(Wrapping<u64>);
impl LibTracker {
    pub(crate) fn new(start: u16) -> Self {
        Self(Wrapping(start as u64))
    }
    pub(crate) fn advance(&mut self, amount: u16) -> Self {
        let old = self.0;
        self.0 = self.0 + Wrapping(amount as u64);
        LibTracker(old)
    }
}
/// Collection of connection tracking, but user-given and library generated
#[derive(Debug, Copy, Clone)]
pub struct ConnTracker {
    /// Optional tracker set by the user
    pub user: Option<UserTracker>,
    /// library generated tracker. Unique and non-repeating
    pub(crate) lib: LibTracker,
}

impl PartialEq for ConnTracker {
    fn eq(&self, other: &Self) -> bool {
        self.lib == other.lib
    }
}
impl Eq for ConnTracker {}

/// Connection to an Authentication Server
#[derive(Debug, Copy, Clone)]
pub struct AuthSrvConn(pub ConnTracker);
/// Connection to a service
#[derive(Debug, Copy, Clone)]
pub struct ServiceConn(pub ConnTracker);

/*
 * TODO: only on Thread{Local,Safe}::Connection oriented flows
/// The connection, as seen from a user of libFenrir
#[derive(Debug)]
pub struct Conn {
    pub(crate) queue: ::async_channel::Sender<worker::Work>,
    pub(crate) tracker: ConnTracker,
}

impl Conn {
    /// Queue some data to be sent in this connection
    // TODO: send_and_wait, that wait for recipient ACK
    pub async fn send(&mut self, stream: stream::ID, data: Vec<u8>) {
        use crate::inner::worker::Work;
        let _ = self
            .queue
            .send(Work::UserSend((self.tracker.lib, stream, data)))
            .await;
    }
    /// Get the library tracking id
    pub fn tracker(&self) -> ConnTracker {
        self.tracker
    }
}
*/

/// Role: track the connection direction
///
/// The Role is used to select the correct secrets, and track the direction
/// of the connection
/// * Server: Conn is Incoming
/// * Client: Conn is Outgoing
#[derive(Debug, Copy, Clone)]
#[repr(u8)]
pub enum Role {
    /// Server: we receive the connection
    Server = 0,
    /// Client: we initate the connection
    Client,
}

#[derive(Debug)]
enum TimerKind {
    None,
    SendData(::tokio::time::Instant),
    Keepalive(::tokio::time::Instant),
}

pub(crate) enum Enqueue {
    TimerWait,
    Immediate(::tokio::time::Instant),
}

/// A single connection and its data
#[derive(Debug)]
pub(crate) struct Connection {
    /// Receiving Conn ID
    pub(crate) id_recv: IDRecv,
    /// Sending Conn ID
    pub(crate) id_send: IDSend,
    /// User-managed id to track this connection
    /// the user can set this to better track this connection
    pub(crate) user_tracker: Option<UserTracker>,
    pub(crate) lib_tracker: LibTracker,
    /// Sending address
    pub(crate) send_addr: UdpClient,
    /// The main hkdf used for all secrets in this connection
    hkdf: Hkdf,
    /// Cipher for decrypting data
    pub(crate) cipher_recv: CipherRecv,
    /// Cipher for encrypting data
    pub(crate) cipher_send: CipherSend,
    mtu: usize,
    next_timer: TimerKind,
    /// send queue for each Stream
    send_queue: BTreeMap<stream::ID, stream::SendTracker>,
    last_stream_sent: stream::ID,
    /// receive queue for each Stream
    recv_queue: BTreeMap<stream::ID, stream::Stream>,
    streams_ready: VecDeque<stream::ID>,
}

impl Connection {
    pub(crate) fn new(
        hkdf: Hkdf,
        cipher: sym::Kind,
        role: Role,
        rand: &Random,
    ) -> Self {
        let (secret_recv, secret_send) = match role {
            Role::Server => {
                (hkdf.get_secret(b"to_server"), hkdf.get_secret(b"to_client"))
            }
            Role::Client => {
                (hkdf.get_secret(b"to_client"), hkdf.get_secret(b"to_server"))
            }
        };
        let cipher_recv = CipherRecv::new(cipher, secret_recv);
        let cipher_send = CipherSend::new(cipher, secret_send, rand);

        use ::std::net::{IpAddr, Ipv4Addr, SocketAddr};
        Self {
            id_recv: IDRecv(ID::Handshake),
            id_send: IDSend(ID::Handshake),
            user_tracker: None,
            lib_tracker: LibTracker::new(0),
            // will be overwritten
            send_addr: UdpClient(SocketAddr::new(
                IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1)),
                31337,
            )),
            hkdf,
            cipher_recv,
            cipher_send,
            mtu: 1200,
            next_timer: TimerKind::None,
            send_queue: BTreeMap::new(),
            last_stream_sent: stream::ID(0),
            recv_queue: BTreeMap::new(),
            streams_ready: VecDeque::with_capacity(4),
        }
    }
    pub(crate) fn get_data(&mut self) -> Option<Vec<(stream::ID, Vec<u8>)>> {
        if self.streams_ready.is_empty() {
            return None;
        }
        let ret_len = self.streams_ready.len();
        let mut ret = Vec::with_capacity(ret_len);
        while let Some(stream_id) = self.streams_ready.pop_front() {
            let stream = match self.recv_queue.get_mut(&stream_id) {
                Some(stream) => stream,
                None => continue,
            };
            let data = stream.get(); // FIXME
            ret.push((stream_id, data.1));
        }
        Some(ret)
    }
    pub(crate) fn recv(
        &mut self,
        mut udp: crate::RawUdp,
    ) -> Result<StreamData, Error> {
        let mut data = &mut udp.data[ID::len()..];
        let aad = sym::AAD(&[]);
        self.cipher_recv.decrypt(aad, &mut data)?;
        let mut bytes_parsed = 0;
        let mut chunks = Vec::with_capacity(2);
        loop {
            let chunk = match stream::Chunk::deserialize(&data[bytes_parsed..])
            {
                Ok(chunk) => chunk,
                Err(e) => {
                    return Err(e.into());
                }
            };
            bytes_parsed = bytes_parsed + chunk.len();
            chunks.push(chunk);
            if bytes_parsed == data.len() {
                break;
            }
        }
        let mut data_ready = StreamData::NotReady;
        for chunk in chunks.into_iter() {
            let stream_id = chunk.id;
            let stream = match self.recv_queue.get_mut(&stream_id) {
                Some(stream) => stream,
                None => {
                    ::tracing::debug!("Ignoring chunk for unknown stream::ID");
                    continue;
                }
            };
            match stream.recv(chunk) {
                Ok(status) => {
                    if !self.streams_ready.contains(&stream_id) {
                        self.streams_ready.push_back(stream_id);
                    }
                    data_ready = data_ready | status;
                }
                Err(e) => ::tracing::debug!("stream: {:?}: {:?}", stream_id, e),
            }
        }
        Ok(data_ready)
    }
    pub(crate) fn enqueue(
        &mut self,
        stream: stream::ID,
        data: Vec<u8>,
    ) -> Result<Enqueue, Error> {
        let stream = match self.send_queue.get_mut(&stream) {
            None => return Err(Error::NoSuchStream),
            Some(stream) => stream,
        };
        stream.enqueue(data);
        let instant;
        let ret;
        self.next_timer = match self.next_timer {
            TimerKind::None | TimerKind::Keepalive(_) => {
                instant = ::tokio::time::Instant::now();
                ret = Enqueue::Immediate(instant);
                TimerKind::SendData(instant)
            }
            TimerKind::SendData(old_timer) => {
                // There already is some data to be sent
                // wait for this timer,
                // or risk going over max transmission rate
                ret = Enqueue::TimerWait;
                TimerKind::SendData(old_timer)
            }
        };
        Ok(ret)
    }
    pub(crate) fn write_pkt<'a>(
        &mut self,
        raw: &'a mut [u8],
    ) -> Result<&'a [u8], enc::Error> {
        assert!(raw.len() >= self.mtu, "I should have at least 1200 MTU");
        if self.send_queue.len() == 0 {
            return Err(enc::Error::NotEnoughData(0));
        }
        raw[..ID::len()]
            .copy_from_slice(&self.id_send.0.as_u64().to_le_bytes());
        let data_from = ID::len() + self.cipher_send.nonce_len().0;
        let data_max_to = raw.len() - self.cipher_send.tag_len().0;
        let mut chunk_from = data_from;
        let mut available_len = data_max_to - data_from;

        use std::ops::Bound::{Excluded, Included};
        let last_stream = self.last_stream_sent;

        // Loop over our streams, write them to the packet.
        // Notes:
        //  * to avoid starvation, just round-robin them all for now
        //  * we can enqueue multiple times the same stream
        //    This is useful especially for Datagram streams
        'queueloop: {
            for (id, stream) in self
                .send_queue
                .range_mut((Included(last_stream), Included(stream::ID::max())))
            {
                if available_len < stream::Chunk::headers_len() + 1 {
                    break 'queueloop;
                }
                let bytes =
                    stream.serialize(*id, &mut raw[chunk_from..data_max_to]);
                if bytes == 0 {
                    break 'queueloop;
                }
                available_len = available_len - bytes;
                chunk_from = chunk_from + bytes;
                self.last_stream_sent = *id;
            }
            if available_len > 0 {
                for (id, stream) in self.send_queue.range_mut((
                    Included(stream::ID::min()),
                    Excluded(last_stream),
                )) {
                    if available_len < stream::Chunk::headers_len() + 1 {
                        break 'queueloop;
                    }
                    let bytes = stream
                        .serialize(*id, &mut raw[chunk_from..data_max_to]);
                    if bytes == 0 {
                        break 'queueloop;
                    }
                    available_len = available_len - bytes;
                    chunk_from = chunk_from + bytes;
                    self.last_stream_sent = *id;
                }
            }
        }
        if chunk_from == data_from {
            return Err(enc::Error::NotEnoughData(0));
        }
        let data_to = chunk_from + self.cipher_send.tag_len().0;

        // encrypt
        let aad = sym::AAD(&[]);
        match self.cipher_send.encrypt(aad, &mut raw[data_from..data_to]) {
            Ok(_) => Ok(&raw[..data_to]),
            Err(e) => Err(e),
        }
    }
}

pub(crate) struct ConnList {
    thread_id: ThreadTracker,
    connections: Vec<Option<Connection>>,
    user_tracker: BTreeMap<LibTracker, usize>,
    last_tracked: LibTracker,
    /// Bitmap to track which connection ids are used or free
    ids_used: Vec<::bitmaps::Bitmap<1024>>,
}

impl ConnList {
    pub(crate) fn new(thread_id: ThreadTracker) -> Self {
        let mut bitmap_id = ::bitmaps::Bitmap::<1024>::new();
        if thread_id.id == 0 {
            // make sure we don't count the Handshake ID
            bitmap_id.set(0, true);
        }
        const INITIAL_CAP: usize = 128;
        let mut ret = Self {
            thread_id,
            connections: Vec::with_capacity(INITIAL_CAP),
            user_tracker: BTreeMap::new(),
            last_tracked: LibTracker(Wrapping(0)),
            ids_used: vec![bitmap_id],
        };
        ret.connections.resize_with(INITIAL_CAP, || None);
        ret
    }
    pub fn get_id_mut(&mut self, id: ID) -> Result<&mut Connection, Error> {
        let conn_id = match id {
            ID::ID(conn_id) => conn_id,
            ID::Handshake => {
                return Err(Error::NoSuchConnection);
            }
        };
        let id_in_thread: usize =
            (conn_id.get() / (self.thread_id.total as u64)) as usize;
        if let Some(conn) = &mut self.connections[id_in_thread] {
            Ok(conn)
        } else {
            return Err(Error::NoSuchConnection);
        }
    }
    pub fn get_mut(
        &mut self,
        tracker: LibTracker,
    ) -> Result<&mut Connection, Error> {
        let idx = if let Some(idx) = self.user_tracker.get(&tracker) {
            *idx
        } else {
            return Err(Error::NoSuchConnection);
        };
        if let Some(conn) = &mut self.connections[idx] {
            Ok(conn)
        } else {
            return Err(Error::NoSuchConnection);
        }
    }
    pub fn len(&self) -> usize {
        let mut total: usize = 0;
        for bitmap in self.ids_used.iter() {
            total = total + bitmap.len()
        }
        total
    }
    /// Only *Reserve* a connection,
    /// without actually tracking it in self.connections
    pub(crate) fn reserve_and_track<'a>(
        &'a mut self,
        mut conn: Connection,
    ) -> (LibTracker, &'a mut Connection) {
        let (id_conn, id_in_thread) = self.reserve_first_with_idx();
        conn.id_recv = id_conn;
        let tracker = self.get_new_tracker(id_in_thread);
        conn.lib_tracker = tracker;
        self.connections[id_in_thread] = Some(conn);
        (tracker, self.connections[id_in_thread].as_mut().unwrap())
    }
    /// Only *Reserve* a connection,
    /// without actually tracking it in self.connections
    pub(crate) fn reserve_first(&mut self) -> IDRecv {
        self.reserve_first_with_idx().0
    }
    fn reserve_first_with_idx(&mut self) -> (IDRecv, usize) {
        // uhm... bad things are going on here:
        // * id must be initialized, but only because:
        // * rust does not understand that after the `!found` id is always
        //   initialized
        // * `ID::new_u64` is really safe only with >0, but here it always is
        // ...we should probably rewrite it in better, safer rust
        let mut id_in_thread: usize = 0;
        let mut found = false;
        for (i, b) in self.ids_used.iter_mut().enumerate() {
            match b.first_false_index() {
                Some(idx) => {
                    b.set(idx, true);
                    id_in_thread = (i * 1024) + idx;
                    found = true;
                    break;
                }
                None => {}
            }
        }
        if !found {
            let mut new_bitmap = ::bitmaps::Bitmap::<1024>::new();
            new_bitmap.set(0, true);
            id_in_thread = self.ids_used.len() * 1024;
            self.ids_used.push(new_bitmap);
        }
        // make sure we have enough space in self.connections
        let curr_capacity = self.connections.capacity();
        if self.connections.capacity() <= id_in_thread {
            // Fill with "None", assure 64 connections without reallocations
            let multiple = 64 + curr_capacity - 1;
            let new_capacity = multiple - (multiple % curr_capacity);
            self.connections.resize_with(new_capacity, || None);
        }
        // calculate the actual connection ID
        let actual_id = ((id_in_thread as u64) * (self.thread_id.total as u64))
            + (self.thread_id.id as u64);
        let new_id = IDRecv(ID::new_u64(actual_id));
        (new_id, id_in_thread)
    }
    /// NOTE: does NOT check if the connection has been previously reserved!
    pub(crate) fn track(
        &mut self,
        mut conn: Connection,
    ) -> Result<LibTracker, ()> {
        let conn_id = match conn.id_recv {
            IDRecv(ID::Handshake) => {
                return Err(());
            }
            IDRecv(ID::ID(conn_id)) => conn_id,
        };
        let id_in_thread: usize =
            (conn_id.get() / (self.thread_id.total as u64)) as usize;
        let tracker = self.get_new_tracker(id_in_thread);
        conn.lib_tracker = tracker;
        self.connections[id_in_thread] = Some(conn);
        Ok(tracker)
    }
    fn get_new_tracker(&mut self, id_in_thread: usize) -> LibTracker {
        let mut tracker;
        loop {
            tracker = self.last_tracked.advance(self.thread_id.total);
            if self.user_tracker.get(&tracker).is_none() {
                // like, never gonna happen, it's 64 bit
                let _ = self.user_tracker.insert(tracker, id_in_thread);
                break;
            }
        }
        tracker
    }
    pub(crate) fn remove(&mut self, id: IDRecv) {
        if let IDRecv(ID::ID(raw_id)) = id {
            let id_in_thread: usize =
                (raw_id.get() / (self.thread_id.total as u64)) as usize;
            let vec_index = id_in_thread / 1024;
            let bitmask_index = id_in_thread % 1024;
            if let Some(bitmask) = self.ids_used.get_mut(vec_index) {
                bitmask.set(bitmask_index, false);
                self.connections[id_in_thread] = None;
            }
        }
    }
}

/// return wether we already have a connection, we are waiting for one, or you
/// can start one
#[derive(Debug, Clone, Copy)]
pub(crate) enum Reservation {
    /// we already have a connection. use this ID.
    Present(IDSend),
    /// we don't have a connection, but we are waiting for one to be established.
    Waiting,
    /// we have reserved a spot for your connection.
    Reserved,
}

enum MapEntry {
    Present(IDSend),
    Reserved,
}

/// Link the public key of the authentication server to a connection id
/// so that we can reuse that connection to ask for more authentications
///
/// Note that a server can have multiple public keys,
/// and the handshake will only ever verify one.
/// To avoid malicious publication fo keys that are not yours,
/// on connection we:
/// * reserve all public keys of the server
/// * wait for the connection to finish
/// * remove all those reservations, exept the one key that actually succeded
/// While searching, we return a connection ID if just one key is a match
// TODO: can we shard this per-core by hashing the pubkey? or domain? or...???
//  This needs a mutex and it will be our goeal to avoid any synchronization
pub(crate) struct AuthServerConnections {
    conn_map: HashMap<PubKey, MapEntry>,
}

impl AuthServerConnections {
    pub(crate) fn new() -> Self {
        Self {
            conn_map: HashMap::with_capacity(32),
        }
    }
    /// add an ID to the reserved spot,
    /// and unlock the other pubkeys which have not been verified
    pub(crate) fn add(
        &mut self,
        pubkey: &PubKey,
        id: IDSend,
        record: &dnssec::Record,
    ) {
        let _ = self.conn_map.insert(*pubkey, MapEntry::Present(id));
        for (_, pk) in record.public_keys.iter() {
            if pk == pubkey {
                continue;
            }
            let _ = self.conn_map.remove(pk);
        }
    }
    /// remove a dropped connection
    pub(crate) fn remove_reserved(&mut self, record: &dnssec::Record) {
        for (_, pk) in record.public_keys.iter() {
            let _ = self.conn_map.remove(pk);
        }
    }
    /// remove a dropped connection
    pub(crate) fn remove_conn(&mut self, pubkey: &PubKey) {
        let _ = self.conn_map.remove(pubkey);
    }

    /// each dnssec::Record has multiple Pubkeys. reserve and ID for them all.
    /// later on, when `add` is called we will delete
    /// those that have not actually benn used
    pub(crate) fn get_or_reserve(
        &mut self,
        record: &dnssec::Record,
    ) -> Reservation {
        for (_, pk) in record.public_keys.iter() {
            match self.conn_map.get(pk) {
                None => {}
                Some(MapEntry::Reserved) => return Reservation::Waiting,
                Some(MapEntry::Present(id)) => {
                    return Reservation::Present(id.clone())
                }
            }
        }
        for (_, pk) in record.public_keys.iter() {
            let _ = self.conn_map.insert(*pk, MapEntry::Reserved);
        }
        Reservation::Reserved
    }
}
