//! Worker thread implementation
use crate::{
    auth::{self, Domain, ServiceID, Token, TokenChecker, UserID},
    config::Config,
    connection::{
        self,
        handshake::{
            self,
            dirsync::{self, DirSync},
            Handshake,
        },
        packet::{self, Packet},
        socket::{UdpClient, UdpServer},
        stream, AuthSrvConn, ConnList, ConnTracker, Connection, IDSend,
        LibTracker, ServiceConn,
    },
    dnssec,
    enc::{
        self,
        asym::{self, KeyID, PrivKey, PubKey},
        hkdf::{self, Hkdf},
        sym, Random, Secret,
    },
    inner::ThreadTracker,
};
use ::std::{collections::VecDeque, sync::Arc, vec::Vec};
/// This worker must be cpu-pinned
use ::tokio::{
    net::UdpSocket,
    sync::{mpsc, oneshot, Mutex},
};

/// Track a raw Udp packet
pub(crate) struct RawUdp {
    pub src: UdpClient,
    pub dst: UdpServer,
    pub data: Vec<u8>,
    pub packet: Packet,
}

pub(crate) struct ConnectInfo {
    pub answer: oneshot::Sender<handshake::tracker::ConnectAnswer>,
    pub resolved: dnssec::Record,
    pub service_id: ServiceID,
    pub domain: Domain,
    // TODO: UserID, Token information
}

/// return to the user the data received from a connection
#[derive(Debug, Clone)]
pub struct ConnData {
    /// Connection tracking information
    pub conn: ConnTracker,
    /// received data, for each stream
    pub data: Vec<(stream::ID, Vec<u8>)>, //FIXME: ChunkOwned
}

/// Connection event. Mostly used to give the data to the user
#[derive(Debug, Clone)]
#[non_exhaustive]
pub enum Event {
    /// Work loop has exited. nothing more to do
    End,
    /// Data from a connection
    Data(ConnData),
}

pub(crate) enum Work {
    /// ask the thread to report to the main thread the total number of
    /// connections present
    CountConnections(oneshot::Sender<usize>),
    Connect(ConnectInfo),
    DropHandshake(KeyID),
    Recv(RawUdp),
    UserSend((LibTracker, stream::ID, Vec<u8>)),
    SendData((LibTracker, ::tokio::time::Instant)),
}

/// Actual worker implementation.
#[allow(missing_debug_implementations)]
pub struct Worker {
    cfg: Config,
    thread_id: ThreadTracker,
    // PERF: rand uses syscalls. how to do that async?
    rand: Random,
    stop_working: crate::StopWorkingRecvCh,
    token_check: Option<Arc<Mutex<TokenChecker>>>,
    sockets: Vec<Arc<UdpSocket>>,
    queue: ::async_channel::Receiver<Work>,
    queue_sender: ::async_channel::Sender<Work>,
    queue_timeouts_recv: mpsc::UnboundedReceiver<Work>,
    queue_timeouts_send: mpsc::UnboundedSender<Work>,
    thread_channels: Vec<::async_channel::Sender<Work>>,
    connections: ConnList,
    // connectsion untracker by the user. (users still needs to get(..) them)
    untracked_connections: VecDeque<LibTracker>,
    handshakes: handshake::Tracker,
    work_timers: super::Timers,
}

#[allow(unsafe_code)]
unsafe impl Send for Worker {}

impl Worker {
    pub(crate) async fn new(
        mut cfg: Config,
        thread_id: ThreadTracker,
        stop_working: crate::StopWorkingRecvCh,
        token_check: Option<Arc<Mutex<TokenChecker>>>,
        sockets: Vec<Arc<UdpSocket>>,
        queue: ::async_channel::Receiver<Work>,
        queue_sender: ::async_channel::Sender<Work>,
    ) -> ::std::io::Result<Self> {
        let (queue_timeouts_send, queue_timeouts_recv) =
            mpsc::unbounded_channel();
        let mut handshakes = handshake::Tracker::new(
            thread_id,
            cfg.ciphers.clone(),
            cfg.key_exchanges.clone(),
        );
        let mut server_keys = Vec::new();
        // make sure the keys are no longer in the config
        ::core::mem::swap(&mut server_keys, &mut cfg.server_keys);
        for k in server_keys.into_iter() {
            if handshakes.add_server_key(k.id, k.priv_key).is_err() {
                return Err(::std::io::Error::new(
                    ::std::io::ErrorKind::AlreadyExists,
                    "You can't use the same KeyID for multiple keys",
                ));
            }
        }
        for srv in cfg.servers.iter() {
            if handshakes.add_server_domain(&srv.fqdn, &srv.keys).is_err() {
                return Err(::std::io::Error::new(
                    ::std::io::ErrorKind::NotFound,
                    "Specified a KeyID that we don't have",
                ));
            }
        }

        Ok(Self {
            cfg,
            thread_id,
            rand: Random::new(),
            stop_working,
            token_check,
            sockets,
            queue,
            queue_sender,
            queue_timeouts_recv,
            queue_timeouts_send,
            thread_channels: Vec::new(),
            connections: ConnList::new(thread_id),
            untracked_connections: VecDeque::with_capacity(8),
            handshakes,
            work_timers: super::Timers::new(),
        })
    }
    /// return a handle to the worker that you can use to send data
    /// The handle will enqueue work in the main worker and is thread-local safe
    ///
    /// While this does not require `&mut` on the `Worker`, everything
    /// will be put in the work queue,
    /// So you might have less immediate results in a few cases
    pub fn handle(&self) -> Handle {
        Handle {
            queue: self.queue_sender.clone(),
        }
    }
    /// change the UserTracker in the connection
    ///
    /// This is `unsafe` because you will be responsible for manually updating
    /// any copy of the `ConnTracker` you might have cloned around
    #[allow(unsafe_code)]
    pub unsafe fn set_connection_tracker(
        &mut self,
        tracker: ConnTracker,
        new_id: connection::UserTracker,
    ) -> Result<ConnTracker, crate::Error> {
        let conn = self.connections.get_mut(tracker.lib)?;
        conn.user_tracker = Some(new_id);
        Ok(ConnTracker {
            lib: tracker.lib,
            user: Some(new_id),
        })
    }
    /// Enqueue data to send
    pub fn send(
        &mut self,
        tracker: LibTracker,
        stream: stream::ID,
        data: Vec<u8>,
    ) -> Result<(), crate::Error> {
        let conn = self.connections.get_mut(tracker)?;
        conn.enqueue(stream, data)?;
        Ok(())
    }
    /// Returns new connections, if any
    ///
    /// You can provide an optional tracker, different from the library tracker.
    ///
    /// Differently from the library tracker, you can change this later on,
    /// but you will be responsible to change it on every `ConnTracker`
    /// you might have cloned elsewhere
    pub fn try_get_connection(
        &mut self,
        tracker: Option<connection::UserTracker>,
    ) -> Option<ConnTracker> {
        let ret_tracker = ConnTracker {
            lib: self.untracked_connections.pop_front()?,
            user: None,
        };
        match tracker {
            Some(tracker) => {
                #[allow(unsafe_code)]
                match unsafe {
                    self.set_connection_tracker(ret_tracker, tracker)
                } {
                    Ok(tracker) => Some(tracker),
                    Err(_) => {
                        // we had a connection, but it expired before the user
                        // remembered to get it. Just remove it from the queue.
                        None
                    }
                }
            }
            None => Some(ret_tracker),
        }
    }

    /// Continuously loop and process work as needed
    pub async fn work_loop(&mut self) -> Result<Event, crate::Error> {
        'mainloop: loop {
            let next_timer = self.work_timers.get_next();
            ::tokio::pin!(next_timer);
            let work = ::tokio::select! {
                tell_stopped = self.stop_working.recv() => {
                    if let Ok(stop_ch) = tell_stopped {
                        let _ = stop_ch.send(
                             crate::StopWorking::WorkerStopped).await;
                    }
                    break;
                }
                () = &mut next_timer =>  {
                    let work_list = self.work_timers.get_work();
                    for w in work_list.into_iter() {
                        let _ = self.queue_sender.send(w).await;
                    }
                    continue 'mainloop;
                }
                maybe_timeout = self.queue.recv() => {
                    match maybe_timeout {
                        Ok(work) => work,
                        Err(_) => break,
                    }
                }
                maybe_work = self.queue.recv() => {
                    match maybe_work {
                        Ok(work) => work,
                        Err(_) => break,
                    }
                }
            };
            match work {
                Work::CountConnections(sender) => {
                    let conn_num = self.connections.len();
                    let _ = sender.send(conn_num);
                }
                Work::Connect(conn_info) => {
                    // PERF: geolocation

                    // Find the first destination with:
                    // * UDP port
                    // * a coherent pubkey/key exchange.
                    let destination =
                        conn_info.resolved.addresses.iter().find_map(|addr| {
                            if addr.port.is_none()
                                || addr
                                    .handshake_ids
                                    .iter()
                                    .find(|h_srv| {
                                        self.cfg.handshakes.contains(h_srv)
                                    })
                                    .is_none()
                            {
                                // skip servers with no corresponding
                                // handshake types or no udp port
                                return None;
                            }

                            // make sure this server has a public key
                            // that supports one of the key exchanges that
                            // *we* support
                            for idx in addr.public_key_idx.iter() {
                                // for each key,
                                // get all the possible key exchanges
                                let key_supported_k_x =
                                    conn_info.resolved.public_keys
                                        [idx.0 as usize]
                                        .1
                                        .kind()
                                        .key_exchanges();
                                // consider only the key exchanges allowed
                                // in the dnssec record
                                let filtered_key_exchanges: Vec<
                                    asym::KeyExchangeKind,
                                > = key_supported_k_x
                                    .into_iter()
                                    .filter(|k_x| {
                                        conn_info
                                            .resolved
                                            .key_exchanges
                                            .contains(k_x)
                                    })
                                    .copied()
                                    .collect();

                                // finally make sure that we support those
                                match self.cfg.key_exchanges.iter().find(|x| {
                                    filtered_key_exchanges.contains(x)
                                }) {
                                    Some(exchange) => {
                                        return Some((
                                            addr,
                                            conn_info.resolved.public_keys
                                                [idx.0 as usize],
                                            exchange.clone(),
                                        ))
                                    }
                                    None => return None,
                                }
                            }
                            return None;
                        });
                    let (addr, key, exchange) = match destination {
                        Some((addr, key, exchange)) => (addr, key, exchange),
                        None => {
                            let _ =
                                conn_info
                                    .answer
                                    .send(Err(crate::Error::Resolution(
                                    "No selectable address and key combination"
                                        .to_owned(),
                                )));
                            continue 'mainloop;
                        }
                    };
                    let hkdf_selected = match hkdf::client_select_hkdf(
                        &self.cfg,
                        &conn_info.resolved.hkdfs,
                    ) {
                        Some(hkdf_selected) => hkdf_selected,
                        None => {
                            let _ = conn_info.answer.send(Err(
                                handshake::Error::Negotiation.into(),
                            ));
                            continue 'mainloop;
                        }
                    };
                    let cipher_selected = match sym::client_select_cipher(
                        &self.cfg,
                        &conn_info.resolved.ciphers,
                    ) {
                        Some(cipher_selected) => cipher_selected,
                        None => {
                            let _ = conn_info.answer.send(Err(
                                handshake::Error::Negotiation.into(),
                            ));
                            continue 'mainloop;
                        }
                    };

                    let (priv_key, pub_key) =
                        match exchange.new_keypair(&self.rand) {
                            Ok(pair) => pair,
                            Err(_) => {
                                ::tracing::error!("Failed to generate keys");
                                let _ = conn_info.answer.send(Err(
                                    handshake::Error::KeyGeneration.into(),
                                ));
                                continue 'mainloop;
                            }
                        };
                    let hkdf;

                    if let PubKey::Exchange(srv_pub) = key.1 {
                        let secret =
                            match priv_key.key_exchange(exchange, srv_pub) {
                                Ok(secret) => secret,
                                Err(_) => {
                                    ::tracing::warn!(
                                        "Could not run the key exchange"
                                    );
                                    let _ = conn_info.answer.send(Err(
                                        handshake::Error::Negotiation.into(),
                                    ));
                                    continue 'mainloop;
                                }
                            };
                        hkdf = Hkdf::new(hkdf_selected, b"fenrir", secret);
                    } else {
                        // crate::dnssec already verifies that the keys
                        // listed in dnssec::Record.addresses.public_key_idx
                        // are PubKey::Exchange
                        unreachable!()
                    }
                    let mut conn = Connection::new(
                        hkdf,
                        cipher_selected,
                        connection::Role::Client,
                        &self.rand,
                    );
                    let dest = UdpClient(addr.as_sockaddr().unwrap());
                    conn.send_addr = dest;

                    let auth_recv_id = self.connections.reserve_first();
                    let service_conn_id = self.connections.reserve_first();
                    conn.id_recv = auth_recv_id;
                    let (client_key_id, hshake) = match self
                        .handshakes
                        .add_client(
                            PrivKey::Exchange(priv_key),
                            PubKey::Exchange(pub_key),
                            conn_info.service_id,
                            service_conn_id,
                            conn,
                            conn_info.answer,
                            key.0,
                        ) {
                        Ok((client_key_id, hshake)) => (client_key_id, hshake),
                        Err(answer) => {
                            ::tracing::warn!("Too many client handshakes");
                            let _ = answer.send(Err(
                                handshake::Error::TooManyClientHandshakes
                                    .into(),
                            ));
                            continue 'mainloop;
                        }
                    };

                    // build request
                    let auth_info = dirsync::req::AuthInfo {
                        user: UserID::new_anonymous(),
                        token: Token::new_anonymous(&self.rand),
                        service_id: conn_info.service_id,
                        domain: conn_info.domain,
                    };
                    let req_data = dirsync::req::Data {
                        nonce: dirsync::Nonce::new(&self.rand),
                        client_key_id,
                        id: auth_recv_id.0, //FIXME: is zero
                        auth: auth_info,
                    };
                    let req = dirsync::req::Req {
                        key_id: key.0,
                        exchange,
                        hkdf: hkdf_selected,
                        cipher: cipher_selected,
                        exchange_key: pub_key,
                        data: dirsync::req::State::ClearText(req_data),
                    };
                    let encrypt_start =
                        connection::ID::len() + req.encrypted_offset();
                    let encrypt_end = encrypt_start
                        + req.encrypted_length(
                            cipher_selected.nonce_len(),
                            cipher_selected.tag_len(),
                        );
                    let h_req = Handshake::new(handshake::Data::DirSync(
                        DirSync::Req(req),
                    ));
                    let packet = Packet {
                        id: connection::ID::Handshake,
                        data: packet::Data::Handshake(h_req),
                    };

                    let tot_len = packet.len(
                        cipher_selected.nonce_len(),
                        cipher_selected.tag_len(),
                    );
                    let mut raw = Vec::<u8>::with_capacity(tot_len);
                    raw.resize(tot_len, 0);
                    packet.serialize(
                        cipher_selected.nonce_len(),
                        cipher_selected.tag_len(),
                        &mut raw[..],
                    );
                    // encrypt
                    if let Err(e) = hshake.connection.cipher_send.encrypt(
                        sym::AAD(&[]),
                        &mut raw[encrypt_start..encrypt_end],
                    ) {
                        ::tracing::error!("Can't encrypt DirSync Request");
                        if let Some(client) =
                            self.handshakes.remove_client(client_key_id)
                        {
                            let _ = client.answer.send(Err(e.into()));
                        };
                        continue 'mainloop;
                    }

                    // send always from the first socket
                    // FIXME: select based on routing table
                    let sender = self.sockets[0].local_addr().unwrap();

                    // start the timeout right before sending the packet
                    let time_drop = self.work_timers.add(
                        ::tokio::time::Duration::from_secs(10),
                        Work::DropHandshake(client_key_id),
                    );
                    hshake.timeout = Some(time_drop);

                    // send packet
                    self.send_packet(raw, dest, UdpServer(sender)).await;

                    continue 'mainloop;
                }
                Work::DropHandshake(key_id) => {
                    if let Some(connections) =
                        self.handshakes.timeout_client(key_id)
                    {
                        for conn_id in connections.into_iter() {
                            self.connections.remove(conn_id);
                        }
                    };
                }
                Work::Recv(pkt) => match self.recv(pkt).await {
                    Ok(event) => return Ok(event),
                    Err(_) => continue 'mainloop,
                },
                Work::UserSend((tracker, stream, data)) => {
                    let conn = match self.connections.get_mut(tracker) {
                        Ok(conn) => conn,
                        Err(_) => continue 'mainloop,
                    };
                    use connection::Enqueue;
                    if let Ok(enqueued) = conn.enqueue(stream, data) {
                        match enqueued {
                            Enqueue::Immediate(instant) => {
                                let _ = self
                                    .queue_sender
                                    .send(Work::SendData((tracker, instant)))
                                    .await;
                            }
                            Enqueue::TimerWait => {}
                        }
                    }
                }
                Work::SendData((tracker, instant)) => {
                    // make sure we don't process events before they are
                    // actually needed.
                    // This is basically busy waiting with extra steps,
                    // but we don't want to spawn lots of timers and
                    // we don't really have a fine-grained sleep that is
                    // multiplatform
                    let now = ::tokio::time::Instant::now();
                    if instant <= now {
                        let _ = self
                            .queue_sender
                            .send(Work::SendData((tracker, instant)))
                            .await;
                        continue;
                    }

                    let mut raw: Vec<u8> = Vec::with_capacity(1200);
                    raw.resize(raw.capacity(), 0);
                    let conn = match self.connections.get_mut(tracker) {
                        Ok(conn) => conn,
                        Err(_) => continue,
                    };
                    let pkt = match conn.write_pkt(&mut raw) {
                        Ok(pkt) => pkt,
                        Err(enc::Error::NotEnoughData(0)) => continue,
                        Err(e) => {
                            ::tracing::error!("Packet generation: {:?}", e);
                            continue;
                        }
                    };
                    let dest = conn.send_addr;
                    let src = UdpServer(self.sockets[0].local_addr().unwrap());
                    let len = pkt.len();
                    raw.truncate(len);
                    let _ = self.send_packet(raw, dest, src);
                }
            }
        }
        Ok(Event::End)
    }
    /// Read and do stuff with the raw udp packet
    async fn recv(&mut self, mut udp: RawUdp) -> Result<Event, ()> {
        if udp.packet.id.is_handshake() {
            let handshake = match Handshake::deserialize(
                &udp.data[connection::ID::len()..],
            ) {
                Ok(handshake) => handshake,
                Err(e) => {
                    ::tracing::debug!("Handshake parsing: {}", e);
                    return Err(());
                }
            };
            let action = match self.handshakes.recv_handshake(
                handshake,
                &mut udp.data[connection::ID::len()..],
            ) {
                Ok(action) => action,
                Err(err) => {
                    ::tracing::debug!("Handshake recv error {}", err);
                    return Err(());
                }
            };
            self.recv_handshake(udp, action).await;
            Err(())
        } else {
            self.recv_packet(udp)
        }
    }
    /// Receive a non-handshake packet
    fn recv_packet(&mut self, udp: RawUdp) -> Result<Event, ()> {
        let conn = match self.connections.get_id_mut(udp.packet.id) {
            Ok(conn) => conn,
            Err(_) => return Err(()),
        };
        match conn.recv(udp) {
            Ok(stream::StreamData::NotReady) => Err(()),
            Ok(stream::StreamData::Ready) => Ok(Event::Data(ConnData {
                conn: ConnTracker {
                    user: conn.user_tracker,
                    lib: conn.lib_tracker,
                },
                data: conn.get_data().unwrap(),
            })),
            Err(e) => {
                ::tracing::trace!("Conn Recv: {:?}", e.to_string());
                Err(())
            }
        }
    }
    /// Receive an handshake packet
    async fn recv_handshake(&mut self, udp: RawUdp, action: handshake::Action) {
        match action {
            handshake::Action::AuthNeeded(authinfo) => {
                let req;
                if let handshake::Data::DirSync(DirSync::Req(r)) =
                    authinfo.handshake.data
                {
                    req = r;
                } else {
                    ::tracing::error!("AuthInfo on non DS::Req");
                    return;
                }
                let req_data = match req.data {
                    dirsync::req::State::ClearText(req_data) => req_data,
                    _ => {
                        ::tracing::error!("AuthNeeded: expected ClearText");
                        assert!(false, "AuthNeeded: unreachable");
                        return;
                    }
                };
                // FIXME: This part can take a while,
                // we should just spawn it probably
                let maybe_auth_check = {
                    match &self.token_check {
                        None => {
                            if req_data.auth.user == auth::USERID_ANONYMOUS {
                                Ok(true)
                            } else {
                                Ok(false)
                            }
                        }
                        Some(token_check) => {
                            let tk_check = token_check.lock().await;
                            tk_check(
                                req_data.auth.user,
                                req_data.auth.token,
                                req_data.auth.service_id,
                                req_data.auth.domain,
                            )
                            .await
                        }
                    }
                };
                let is_authenticated = match maybe_auth_check {
                    Ok(is_authenticated) => is_authenticated,
                    Err(_) => {
                        ::tracing::error!("error in token auth");
                        // TODO: retry?
                        return;
                    }
                };
                if !is_authenticated {
                    ::tracing::warn!(
                        "Wrong authentication for user {:?}",
                        req_data.auth.user
                    );
                    // TODO: error response
                    return;
                }
                // Client has correctly authenticated
                // TODO: contact the service, get the key and
                // connection ID
                let srv_conn_id = connection::ID::new_rand(&self.rand);
                let srv_secret = Secret::new_rand(&self.rand);
                let head_len = req.cipher.nonce_len();
                let tag_len = req.cipher.tag_len();

                let mut auth_conn = Connection::new(
                    authinfo.hkdf,
                    req.cipher,
                    connection::Role::Server,
                    &self.rand,
                );
                auth_conn.id_send = IDSend(req_data.id);
                auth_conn.send_addr = udp.src;
                // track connection
                let auth_id_recv = self.connections.reserve_first();
                auth_conn.id_recv = auth_id_recv;
                let (tracker, auth_conn) =
                    self.connections.reserve_and_track(auth_conn);
                self.untracked_connections.push_back(tracker);

                let resp_data = dirsync::resp::Data {
                    client_nonce: req_data.nonce,
                    id: auth_conn.id_recv.0,
                    service_connection_id: srv_conn_id,
                    service_key: srv_secret,
                };
                use crate::enc::sym::AAD;
                // no aad for now
                let aad = AAD(&mut []);

                let resp = dirsync::resp::Resp {
                    client_key_id: req_data.client_key_id,
                    data: dirsync::resp::State::ClearText(resp_data),
                };
                let encrypt_from =
                    connection::ID::len() + resp.encrypted_offset();
                let encrypt_until =
                    encrypt_from + resp.encrypted_length(head_len, tag_len);
                let resp_handshake = Handshake::new(handshake::Data::DirSync(
                    DirSync::Resp(resp),
                ));
                let packet = Packet {
                    id: connection::ID::new_handshake(),
                    data: packet::Data::Handshake(resp_handshake),
                };
                let tot_len = packet.len(head_len, tag_len);
                let mut raw_out = Vec::<u8>::with_capacity(tot_len);
                raw_out.resize(tot_len, 0);
                packet.serialize(head_len, tag_len, &mut raw_out);

                if let Err(e) = auth_conn
                    .cipher_send
                    .encrypt(aad, &mut raw_out[encrypt_from..encrypt_until])
                {
                    ::tracing::error!("can't encrypt: {:?}", e);
                    return;
                }
                self.send_packet(raw_out, udp.src, udp.dst).await;
            }
            handshake::Action::ClientConnect(cci) => {
                self.work_timers.remove(cci.old_timeout);
                let ds_resp;
                if let handshake::Data::DirSync(DirSync::Resp(resp)) =
                    cci.handshake.data
                {
                    ds_resp = resp;
                } else {
                    ::tracing::error!("ClientConnect on non DS::Resp");
                    return;
                }
                // track connection
                let resp_data;
                if let dirsync::resp::State::ClearText(r_data) = ds_resp.data {
                    resp_data = r_data;
                } else {
                    ::tracing::error!(
                        "ClientConnect on non DS::Resp::ClearText"
                    );
                    unreachable!();
                }
                let auth_id_send = IDSend(resp_data.id);
                let mut conn = cci.connection;
                conn.id_send = auth_id_send;
                let id_recv = conn.id_recv;
                let cipher = conn.cipher_recv.kind();
                // track the connection to the authentication server
                let track_auth_conn = match self.connections.track(conn) {
                    Ok(track_auth_conn) => track_auth_conn,
                    Err(_) => {
                        ::tracing::error!(
                            "Could not track new auth srv connection"
                        );
                        self.connections.remove(id_recv);
                        // FIXME: proper connection closing
                        let _ = cci.answer.send(Err(
                            handshake::Error::InternalTracking.into(),
                        ));
                        return;
                    }
                };
                let authsrv_conn = AuthSrvConn(ConnTracker {
                    lib: track_auth_conn,
                    user: None,
                });
                let mut service_conn = None;
                if cci.service_id != auth::SERVICEID_AUTH {
                    // create and track the connection to the service
                    // SECURITY: xor with secrets
                    //FIXME: the Secret should be XORed with the client
                    // stored secret (if any)
                    let hkdf = Hkdf::new(
                        hkdf::Kind::Sha3,
                        cci.service_id.as_bytes(),
                        resp_data.service_key,
                    );
                    let mut service_connection = Connection::new(
                        hkdf,
                        cipher,
                        connection::Role::Client,
                        &self.rand,
                    );
                    service_connection.id_recv = cci.service_connection_id;
                    service_connection.id_send =
                        IDSend(resp_data.service_connection_id);
                    let track_serv_conn =
                        match self.connections.track(service_connection) {
                            Ok(track_serv_conn) => track_serv_conn,
                            Err(_) => {
                                ::tracing::error!(
                                    "Could not track new service connection"
                                );
                                self.connections
                                    .remove(cci.service_connection_id);
                                // FIXME: proper connection closing
                                // FIXME: drop auth srv connection if we just
                                // established it
                                let _ = cci.answer.send(Err(
                                    handshake::Error::InternalTracking.into(),
                                ));
                                return;
                            }
                        };
                    service_conn = Some(ServiceConn(ConnTracker {
                        lib: track_serv_conn,
                        user: None,
                    }));
                }
                let _ = cci.answer.send(Ok(handshake::tracker::ConnectOk {
                    auth_key_id: cci.srv_key_id,
                    auth_id_send,
                    authsrv_conn,
                    service_conn,
                }));
            }
            handshake::Action::Nothing => {}
        };
    }
    async fn send_packet(
        &self,
        data: Vec<u8>,
        client: UdpClient,
        server: UdpServer,
    ) {
        let src_sock = match self
            .sockets
            .iter()
            .find(|&s| s.local_addr().unwrap() == server.0)
        {
            Some(src_sock) => src_sock,
            None => {
                ::tracing::error!(
                    "Can't send packet: Server changed listening ip{}!",
                    server.0
                );
                return;
            }
        };
        let _res = src_sock.send_to(&data, client.0).await;
    }
}

/// Handle to send work asyncronously to the worker
#[derive(Debug, Clone)]
pub struct Handle {
    queue: ::async_channel::Sender<Work>,
}

impl Handle {
    // TODO
    // pub fn send(..)
    // pub fn set_connection_id(..)
    // try_get_new_connections()
}
